unit SliderBars;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Graphics, Controls, ExtCtrls, Spin;

type

  TSliderBarOrientation = (sboHorizontal, sboVertical);
  TMarkerStyle = (msTriangular, msFillet, msRectangular, msNone);

  { TCustomSliderBar }

  TCustomSliderBar = class(TCustomControl)
  private
    FOrientation: TSliderBarOrientation;
    FBorderColor: TColor;
    FMarginColor: TColor;
    FMarkerColor: TColor;
    FMarkerStyle: TMarkerStyle;
    FMinValue: integer;
    FMaxValue: integer;
    FPosition: extended;
    FLeftMarker,
    FTopMarker,
    FRightMarker,
    FBottomMarker: array of TPoint;
    FAssociate: TSpinEdit;
    FAssociateUpdating: boolean;
    FOnChange: TNotifyEvent;
    procedure AssociateChange(Sender: TObject);
    procedure Change;
    procedure ChangeAssociate;
    function GetValue: integer;
    procedure SetAssociate(AValue: TSpinEdit);
    procedure SetBorderColor(AValue: TColor);
    procedure SetMarkerStyle(AValue: TMarkerStyle);
    procedure SetMarginColor(AValue: TColor);
    procedure SetMarkerColor(AValue: TColor);
    procedure SetMaxValue(AValue: integer);
    procedure SetMinValue(AValue: integer);
    procedure SetOrientation(AValue: TSliderBarOrientation);
    procedure SetPosition(AValue: extended);
    procedure SetValue(AValue: integer);
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure MouseDown(Button: TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseMove(Shift: TShiftState; X,Y: Integer); override;
    procedure Resize; override;
  public
    constructor create(AOwner: TComponent); override;
    destructor destroy; override;
  protected
    procedure Paint; override;
    procedure PaintContent(aRect: TRect); virtual;

    property Associate: TSpinEdit read FAssociate write SetAssociate;
    property BorderColor: TColor read FBorderColor write SetBorderColor default clSilver;
    property MarginColor: TColor read FMarginColor write SetMarginColor default clWhite;
    property MarkerColor: TColor read FMarkerColor write SetMarkerColor default clBlack;
    property MarkerStyle: TMarkerStyle read FMarkerStyle write SetMarkerStyle default msTriangular;
    property Orientation: TSliderBarOrientation read FOrientation write SetOrientation default sboHorizontal;
    property Width default 200;
    property Height default 28;

    property MinValue: integer read FMinValue write SetMinValue default 0;
    property MaxValue: integer read FMaxValue write SetMaxValue default 100;
    property Value: integer read GetValue write SetValue default 0;

    property Position: extended read FPosition;
    property TabStop default true;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  TSliderPaint = procedure(Sender: TObject; aRect: TRect) of object;

  { TSliderBar }

  TSliderBar = class(TCustomSliderBar)
  private
    FOnPaint: TSliderPaint;
  protected
    procedure PaintContent(aRect: TRect); override;
  published
    property Align;
    property Anchors;
    property Associate;
    property BorderColor;
    property BorderSpacing;
    property Constraints;
    property Cursor;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Height;
    property MarginColor;
    property MarkerColor;
    property MarkerStyle;
    property MaxValue;
    property MinValue;
    property Orientation;
    property ParentShowHint;
    property PopupMenu;
    property Position;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Value;
    property Visible;
    property Width;
    property OnChange;
    property OnClick;
    property OnContextPopup;
    property OnDockDrop;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnPaint: TSliderPaint read FOnPaint write FOnPaint;
    property OnResize;
    property OnStartDrag;
  end;

  { TColorBar }

  TColorBar = class(TCustomSliderBar)
  private
    FStartColor: TColor;
    FStopColor: TColor;
    procedure SetStartColor(AValue: TColor);
    procedure SetStopColor(AValue: TColor);
  protected
    procedure PaintContent(aRect: TRect); override;
  public
    procedure SetColors(aStart, aStop: TColor);
  published
    property Align;
    property Anchors;
    property Associate;
    property BorderColor;
    property BorderSpacing;
    property Constraints;
    property Cursor;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Height;
    property MarginColor;
    property MarkerColor;
    property MarkerStyle;
    property MaxValue;
    property MinValue;
    property Orientation;
    property ParentShowHint;
    property PopupMenu;
    property Position;
    property ShowHint;
    property StartColor: TColor read FStartColor write SetStartColor;
    property StopColor: TColor read FStopColor write SetStopColor;
    property TabOrder;
    property TabStop;
    property Value;
    property Visible;
    property Width;
    property OnChange;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnResize;
    //property OnStartDock;
    property OnStartDrag;
    //property OnUnDock;
  end;


implementation

uses
  LCLType, LCLIntf;

{ TCustomSliderBar }

constructor TCustomSliderBar.create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csRequiresKeyboardInput];
  FBorderColor := clSilver;
  FMarginColor := clWhite;
  //FMarkerColor := clBlack;
  //FMarkerStyle := msFillet;
  //FOrientation := sboHorizontal;
  //FMinValue := 0;
  FMaxValue := 100;
  SetLength(FTopMarker, 3);
  SetLength(FBottomMarker, 3);
  SetLength(FLeftMarker, 3);
  SetLength(FRightMarker, 3);
  SetBounds(Left, Top, 200, 28);
  TabStop := true;
end;

destructor TCustomSliderBar.destroy;
begin
  if assigned(FAssociate) then
    FAssociate.OnChange := nil;
  inherited;
end;

procedure TCustomSliderBar.AssociateChange(Sender: TObject);
begin
  FAssociateUpdating := true;
  try
    SetValue(Associate.Value);
  finally
    FAssociateUpdating := false;
  end;
end;

procedure TCustomSliderBar.Change;
begin
  ChangeAssociate;
  if assigned(FOnChange) then
    FOnChange(self);
end;

procedure TCustomSliderBar.ChangeAssociate;
begin
  if assigned(Associate) and not FAssociateUpdating then begin
    // don't want associate's OnChange to start loop;
    Associate.OnChange := nil;
    try
      Associate.Value := Value;
    finally
      Associate.OnChange := @AssociateChange;
    end;
  end;
end;

function TCustomSliderBar.GetValue: integer;
begin
  result := round(Position*(MaxValue - MinValue) + MinValue);
end;

procedure TCustomSliderBar.KeyDown(var Key: Word; Shift: TShiftState);
var
  step: integer;
begin
  step := 1;
  if ssShift in Shift then
    step := 5;
  if (Key = VK_RIGHT) or (Key = VK_UP) then
    SetValue(GetValue + step)
  else if (Key = VK_LEFT) or (Key = VK_DOWN) then
    SetValue(GetValue - step)
  else begin
    inherited KeyDown(Key, Shift);
    exit;
  end;
  Key := 0;
end;

procedure TCustomSliderBar.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  if (AComponent = FAssociate) and (Operation = opRemove) then
    FAssociate := nil;
  inherited Notification(AComponent, Operation);
end;

procedure TCustomSliderBar.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if Button = mbLeft then begin
     if TabStop then SetFocus;
    if Orientation = sboHorizontal then begin
      if x < 2 then
        x := 2
      else if x > width - 2 then
        x := width-2;
      SetPosition( (x - 2)/(width - 4) );
    end
    else {if Orientation = sboVertical then} begin
      if y < 2 then
        y := 2
      else if y > height - 2 then
        y := height-2;
      SetPosition( 1 - (y - 2)/(height - 4) );
    end
  end;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TCustomSliderBar.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if shift = [ssLeft] then begin
    if Orientation = sboHorizontal then begin
      if x < 2 then
        x := 2
      else if x > width - 2 then
        x := width-2;
      SetPosition( (x - 2)/(width - 4) );
    end
    else {if Orientation = sboVertical then} begin
      if y < 2 then
        y := 2
      else if y > height - 2 then
        y := height-2;
      SetPosition( 1 - (y - 2)/(height - 4) );
    end
  end;
  inherited MouseMove(Shift, X, Y);
end;

procedure TCustomSliderBar.SetAssociate(AValue: TSpinEdit);
begin
  if FAssociate=AValue then Exit;
  if assigned(FAssociate) then
    FAssociate.OnChange := nil;
  FAssociate:=AValue;
  if assigned(FAssociate) then begin
    FAssociate.MaxValue := MaxValue;
    FAssociate.MinValue := MinValue;
    ChangeAssociate;
  end;
end;

procedure TCustomSliderBar.SetBorderColor(AValue: TColor);
begin
  if FBorderColor=AValue then Exit;
  FBorderColor:=AValue;
  Invalidate;
end;

procedure TCustomSliderBar.SetMarginColor(AValue: TColor);
begin
  if FMarginColor=AValue then Exit;
  FMarginColor:=AValue;
  Invalidate;
end;

procedure TCustomSliderBar.SetMarkerColor(AValue: TColor);
begin
  if FMarkerColor=AValue then Exit;
  FMarkerColor:=AValue;
  Invalidate;
end;

procedure TCustomSliderBar.SetMaxValue(AValue: integer);
var
  pos: extended;
begin
  if FMaxValue=AValue then Exit;
  pos := position;
  FMaxValue:=AValue;
  if assigned(fAssociate) then
    FAssociate.MaxValue := AValue;
  FPosition := pos;
  Change;
end;

procedure TCustomSliderBar.SetMinValue(AValue: integer);
var
  pos: extended;
begin
  if FMinValue=AValue then Exit;
  pos := position;
  FMinValue:=AValue;
  if assigned(fAssociate) then
    FAssociate.MinValue := AValue;
  FPosition := pos;
  Change;
end;

procedure TCustomSliderBar.SetMarkerStyle(AValue: TMarkerStyle);
begin
  if FMarkerStyle=AValue then Exit;
  FMarkerStyle:=AValue;
  Invalidate;
end;

procedure TCustomSliderBar.SetOrientation(AValue: TSliderBarOrientation);
begin
  if FOrientation=AValue then Exit;
  FOrientation:=AValue;
  Invalidate;
end;

procedure TCustomSliderBar.SetPosition(AValue: extended);
begin
  if FPosition=AValue then Exit;
  if AValue < 0.0 then
    AValue := 0.0
  else if AValue > 1.0 then
    AValue := 1.0;
  FPosition:=AValue;
  Invalidate;
  Change;
end;

procedure TCustomSliderBar.SetValue(AValue: integer);
begin
  SetPosition( (aValue - FMinValue)/(FMaxValue - FMinValue) );
end;

procedure TCustomSliderBar.Resize;
begin
  inherited Resize;
  invalidate;
end;

procedure TCustomSliderBar.Paint;
const
  MarkerSize = 4;
var
  aRect: TRect;
  x, y, y1, y2, x1, x2: integer;
begin
  // First paint the content... done by descendants
  aRect := ClientRect;
  inflateRect(aRect, -2, -2);
  PaintContent(aRect);

  if MarkerStyle <> msNone then begin
    // calculate marker's position
    if orientation = sboHorizontal then
      x := round ( position * (aRect.Right - aRect.left - 1)) + aRect.Left
    else
      y := aRect.bottom - round ( position * (aRect.bottom - aRect.top - 1)) - 1;
  end;

  // paint margin
  inflateRect(aRect, 1, 1);
  with Canvas do begin
    pen.width := 1;
    pen.style := psSolid;
    pen.mode := pmCopy;
    pen.color := MarginColor;
    Frame(aRect);
  end;

  if MarkerStyle <> msNone then begin

    if MarkerStyle <> msRectangular then begin
      if orientation = sboHorizontal then begin
        FTopMarker[2] := Point(x, aRect.Top+MarkerSize);
        FTopMarker[0] := Point(x-MarkerSize, aRect.Top);
        FTopMarker[1] := Point(x+MarkerSize, aRect.Top);

        FBottomMarker[2] := Point(x, aRect.Bottom-MarkerSize-1);
        FBottomMarker[0] := Point(x-MarkerSize, aRect.Bottom-1);
        FBottomMarker[1] := Point(x+MarkerSize, aRect.Bottom-1);
      end
      else begin
        FLeftMarker[2] := Point(aRect.Left+MarkerSize, y);
        FLeftMarker[0] := Point(aRect.left, y-MarkerSize);
        FLeftMarker[1] := Point(aRect.left, y+MarkerSize);

        FRightMarker[2] := Point(aRect.Right-MarkerSize-1, y);
        FRightMarker[0] := Point(aRect.Right-1, y-MarkerSize);
        FRightMarker[1] := Point(aRect.Right-1, y+MarkerSize);
      end;
    end;

    with canvas do begin
      if MarkerStyle <> msTriangular then begin
        // draw margin around marker line
        pen.color := MarginColor;
        if orientation = sboHorizontal then begin
          moveto(x-1, aRect.Top+1);
          lineto(x-1, aRect.Bottom-1);
          moveto(x+1, aRect.Top+1);
          lineto(x+1, aRect.Bottom-1);
        end
        else begin
          moveto(aRect.Left+1, y-1);
          lineto(aRect.Right-1, y-1);
          moveto(aRect.Left+1, y+1);
          lineto(aRect.Right-1, y+1);
        end;
      end;
      if MarkerStyle <> msRectangular then begin
        // draw markers
        pen.color := MarginColor;
        brush.color := MarkerColor;
        if Orientation = sboHorizontal then begin
          Polygon(FTopMarker);
          Polygon(FBottomMarker);
        end
        else begin
          Polygon(FLeftMarker);
          Polygon(FRightMarker);
        end;
      end;

      if MarkerStyle <> msTriangular then begin
        // draw marker line
        pen.color := MarkerColor;
        if orientation = sboHorizontal then begin
          y1 := aRect.Top+1;
          y2 := aRect.Bottom-1;
          if MarkerStyle = msFillet then begin
            inc(y1);
            dec(y2);
          end
          else if MarkerColor = BorderColor then begin
            dec(y1, 2);
            inc(y2, 2);
          end;
          moveto(x, y1);
          lineto(x, y2);
        end
        else {if Orientation = sboVertical then} begin
          x1 := aRect.Left+1;
          x2 := aRect.Right-1;
          if MarkerStyle = msFillet then begin
            inc(x1);
            dec(x2);
          end
          else if MarkerColor = BorderColor then begin
            dec(x1, 2);
            inc(x2, 2);
          end;
          moveto(x1, y);
          lineto(x2, y);
        end;
      end;
    end;
  end;

  with canvas do begin
     // draw frame
    if Focused then
      pen.color := clHighlight
    else
      pen.color := BorderColor;
    aRect := ClientRect;
    Frame(aRect);
  end;
end;

procedure TCustomSliderBar.PaintContent(aRect: TRect);
begin
  // do nothing, its up to the descendants
end;

{ TSliderBar }

procedure TSliderBar.PaintContent(aRect: TRect);
begin
  if assigned(FOnPaint) then
    FOnPaint(Self, aRect);
end;

{ TColorBar }

procedure TColorBar.PaintContent(aRect: TRect);
begin
  if orientation = sboHorizontal then
    Canvas.GradientFill(aRect, FStartColor, FStopColor, gdHorizontal)
  else
    Canvas.GradientFill(aRect, FStopColor, FStartColor, gdVertical)
end;

procedure TColorBar.SetColors(aStart, aStop: TColor);
begin
  FStartColor := aStart;
  FStopColor := aStop;
  Invalidate;
end;

procedure TColorBar.SetStartColor(AValue: TColor);
begin
  if FStartColor=AValue then Exit;
  FStartColor:=AValue;
  Invalidate;
end;

procedure TColorBar.SetStopColor(AValue: TColor);
begin
  if FStopColor=AValue then Exit;
  FStopColor:=AValue;
  Invalidate;
end;

end.

